"""
Copyright 2022 - David Minton, Carlisle Wishard, Jennifer Pouplin, Jake Elliott, & Dana Singh
This file is part of Swiftest.
Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Swiftest. 
If not, see: https://www.gnu.org/licenses. 
"""

import numpy as np
import xarray as xr
"""
Functions that recreate the Swift/Swifter tool programs
"""
def magnitude(ds,x):
    dim = "space"
    ord = None
    return xr.apply_ufunc(
        np.linalg.norm, ds[x].where(~np.isnan(ds[x])), input_core_dims=[[dim]], kwargs={"ord": ord, "axis": -1}, dask="allowed"
    )
        
def wrap_angle(angle):
    """
    Converts angles to be between 0 and 360 degrees.
        
    Parameters
    ----------
    angle : float
        Angle to be converted
    Returns
    -------
    angle : float
        Converted angle
    """
    while np.any(angle >= 360.0 ):
        angle[angle >= 360.0] -= 360.0
    while np.any(angle < 0.0):
        angle[angle < 0.0] += 360.0
    return angle

def follow_swift(ds, ifol=None, nskp=None):
    """
    Emulates the Swift version of tool_follow.f


    Parameters
    ----------
    ds : Dataset containing orbital elements

    Returns
    -------
    fol : Dataset containing only the followed body with angles converted to degrees
    
    Generates a file called follow.out containing 10 columns (angles are all in degrees):
        1 2 3  4    5     6    7    8    9   10
        t,a,e,inc,capom,omega,capm,peri,apo,obar

    """
    fol = None
    if ifol is None:
        intxt = input(' Input the particle number to follow \n')
        ifol = int(intxt)
    print(f"Following particle {ifol}")
    if ifol < 0: # Negative numbers are planets
        fol = ds.where(np.invert(np.isnan(ds['Gmass'])), drop=True)
        fol = fol.where(np.invert(np.isnan(fol['a'])), drop=True) # Remove times where this body doesn't exist (but this also gets rid of the central body)
        fol = fol.isel(id = -ifol - 2)  # Take 1 off for 0-indexed arrays in Python, and take 1 more off because the central body is gone
    elif ifol > 0: # Positive numbers are test particles
        fol = ds.where(np.isnan(ds['Gmass']), drop=True).drop_vars(['Gmass', 'radius'])
        fol = fol.where(np.invert(np.isnan(fol['a'])), drop=True)
        fol = fol.isel(id = ifol - 1)  # Take 1 off for 0-indexed arrays in Python

    if nskp is None:
        intxt = input('Input the print frequency\n')
        nskp = int(intxt)
        
    fol['obar'] = fol['capom'] + fol['omega']
    fol['obar'] = fol['obar'].fillna(0)
    fol['obar'] = wrap_angle(fol['obar'])
    fol['peri'] = fol['a'] * (1.0 - fol['e'])
    fol['apo']  = fol['a'] * (1.0 + fol['e'])

    
    tslice = slice(None, None, nskp)
    try:
        with open('follow.out', 'w') as f:
            print('# 1 2 3  4    5     6    7    8    9   10', file=f)
            print('# t,a,e,inc,capom,omega,capm,peri,apo,obar', file=f)
            for t in fol.isel(time=tslice).time:
                a = fol['a'].sel(time=t).values
                e = fol['e'].sel(time=t).values
                inc = fol['inc'].sel(time=t).values
                capom = fol['capom'].sel(time=t).values
                omega = fol['omega'].sel(time=t).values
                capm = fol['capm'].sel(time=t).values
                peri = fol['peri'].sel(time=t).values
                apo = fol['apo'].sel(time=t).values
                obar = fol['obar'].sel(time=t).values
                print(f"{t.values:15.7e} {a:22.16f} {e:22.16f} {inc:22.16f} {capom:22.16f} {omega:22.16f} {capm:22.16f} {peri:22.16f} {apo:22.16f} {obar:22.16f}", file=f)
                
    except IOError:
        print(f"Error writing to follow.out")
    
    return fol

def danby(M, ecc, accuracy=1e-14):
    """
    Danby's method to solve Kepler's equation.

    Parameters
    ----------
    M : float
        the mean anomaly in radians
    ecc : float
        the eccentricity
    accuracy : float
        the relative accuracy to obtain a solution. Default is 1e-12.

    Returns
    ----------
    E : float
        the eccentric anomaly in radians

    References
    __________
    Danby, J.M.A. 1988. Fundamentals of celestial mechanics. Richmond, Va., U.S.A., Willmann-Bell, 1988. 2nd ed.
    Murray, C.D., Dermott, S.F., 1999. Solar system dynamics, New York, New York. ed, Cambridge University Press.

    """

    def kepler_root(E, ecc, M, deriv):
        """
        The Kepler equation root function.

        Parameters
        ----------
        E : float
            the eccentric anomaly in radians
        ecc : float
            the eccentricity
        M : float
            the mean anomaly in radians
        deriv : int between 0 and 3
            The order of the derivative to compute

        Returns
        ----------
        deriv = 0: E - e * np.sin(E) - M
        deriv = 1: 1 - e * np.cos(E)
        deriv = 2: e * np.sin(E)
        deriv = 3: e * np.cos(E)

        Note: The function will return 0 when E is correct for a given e and M
        """

        if deriv == 0:
            return E - ecc * np.sin(E) - M
        elif deriv == 1:
            return 1.0 - ecc * np.cos(E)
        elif deriv == 2:
            return ecc * np.sin(E)
        elif deriv == 3:
            return ecc * np.cos(E)
        else:
            print(f"deriv={deriv} is undefined")
            return None

    def delta_ij(E, ecc, M, j):
        """
        Danby's intermediate delta functions. This function is recursive for j>1.

        Parameters
        ----------
        E : float
            the eccentric anomaly in radians
        ecc : float
            the eccentricity
        M : float
            the mean anomaly in radians
        j : int between 1 and 3
            The order of the delta function to compute

        Returns
        ----------
        delta_ij value used in Danby's iterative Kepler equation solver

        """
        if j == 1:
            return -kepler_root(E, ecc, M, 0) / kepler_root(E, ecc, M, 1)
        elif j == 2:
            return -kepler_root(E, ecc, M, 0) / (kepler_root(E, ecc, M, 1)
                                                 - delta_ij(E, ecc, M, 1) * kepler_root(E, ecc, M, 2) / 2.0)
        elif j == 3:
            return -kepler_root(E, ecc, M, 0) / (kepler_root(E, ecc, M, 1)
                                                 + delta_ij(E, ecc, M, 1) * kepler_root(E, ecc, M, 2) / 2.0
                                                 + delta_ij(E, ecc, M, 2) ** 2 * kepler_root(E, ecc, M,
                                                                                             3) / 6.0)
        else:
            print(f"j = {j} is not a valid input to the delta_ij function")

    # If this is a circular orbit, just return the mean anomaly as the eccentric anomaly
    if ecc < np.finfo(np.float64).tiny:  # Prevent floating point exception for 0 eccentricity orbits
        return M

    # Initial guess
    k = 0.85
    E = M + np.sign(np.sin(M)) * k * ecc
    MAXLOOPS = 50  # Maximum number of times to iterate before we give up
    for i in range(MAXLOOPS):
        Enew = E + delta_ij(E, ecc, M, 3)
        if np.abs((Enew - E) / E) < accuracy:
            return Enew
        E = Enew

    raise RuntimeError("The danby function did not converge on a solution.")



def el2xv_one(mu, a, ecc, inc, Omega, omega, M):
    """
    Compute osculating orbital elements from relative Cartesian position and velocity
    All angular measures are returned in radians
        If inclination < TINY, longitude of the ascending node is arbitrarily set to 0

        If eccentricity < sqrt(TINY), argument of pericenter is arbitrarily set to 0

          ALGORITHM:  See Fitzpatrick "Principles of Cel. Mech."

    Adapted from Martin Duncan's el2xv.f

    Parameters
    ----------
    mu : float
        Central body gravitational constant
    a : float
        semimajor axis
    ecc : float
        eccentricity
    inc : float
        inclination (degrees)
    Omega : float
        longitude of ascending node (degrees)
    omega : float
        argument of periapsis (degrees)
    M : flaot
        Mean anomaly (degrees)

    Returns
    ----------
    rvec, vvec  : tuple of float vectors

    rvec : (3) float vector
        Cartesian position vector
    vvec : (3) float vector
        Cartesian velocity vector

    """

    if ecc < 0.0:
        print("Error in el2xv! Eccentricity cannot be negative. Setting it to 0")
        e = 0.0
        iorbit_type = "ellipse"
    else:
        e = ecc
        em1 = e - 1.
        if np.abs(em1) < 2 * np.finfo(float).eps:
            iorbit_type = "parabola"
        elif e > 1.0:
            iorbit_type = "hyperbola"
        else:
            iorbit_type = "ellipse"

    def scget(angle):
        """
        Efficiently compute the sine and cosine of an input angle
        Input angle must be in radians

        Adapted from David E. Kaufmann's Swifter routine: orbel_scget.f90
        Adapted from Hal Levison's Swift routine orbel_scget.f

        Parameters
        ----------
        angle : input angle

        Returns
        -------
        sx : sin of angle
        cx : cos of angle

        """
        TWOPI = 2 * np.pi
        nper = int(angle / TWOPI)
        x = angle - nper * TWOPI
        if x < 0.0:
            x += TWOPI

        sx = np.sin(x)
        cx = np.sqrt(1.0 - sx ** 2)
        if x > np.pi / 2.0 and x < 3 * np.pi / 2.0:
            cx = -cx
        return sx, cx

    sip, cip = scget(np.deg2rad(omega))
    so, co = scget(np.deg2rad(Omega))
    si, ci = scget(np.deg2rad(inc))

    d11 = cip * co - sip * so * ci
    d12 = cip * so + sip * co * ci
    d13 = sip * si
    d21 = -sip * co - cip * so * ci
    d22 = -sip * so + cip * co * ci
    d23 = cip * si

    # Get the other quantities depending on orbit type
    if iorbit_type == "ellipse":
        E = danby(np.deg2rad(M),e)
        scap, ccap = scget(E)
        sqe = np.sqrt(1. - e**2)
        sqgma = np.sqrt(mu * a)
        xfac1 = a * (ccap - e)
        xfac2 = a * sqe * scap
        ri = 1. / (a * (1. - e * ccap))
        vfac1 = -ri * sqgma * scap
        vfac2 = ri * sqgma * sqe * ccap
    else:
        print(f"Orbit type: {iorbit_type} not yet implemented")
        xfac1 = 0.0
        xfac2 = 0.0
        vfac1 = 0.0
        vfac2 = 0.0

    rvec = np.array([d11 * xfac1 + d21 * xfac2,
                     d12 * xfac1 + d22 * xfac2,
                     d13 * xfac1 + d23 * xfac2])
    vvec = np.array([d11 * vfac1 + d21 * vfac2,
                     d12 * vfac1 + d22 * vfac2,
                     d13 * vfac1 + d23 * vfac2])

    return rvec, vvec


def el2xv_vec(mu, a, ecc, inc, Omega, omega, M):
    """

    Vectorized call to el2xv_one
    Parameters
    ----------
    mu : float
        Central body gravitational constant
    a : (n) float array
        semimajor axis
    ecc : (n) float array
        eccentricity
    inc : (n) float array
        inclination (degrees)
    Omega : (n) float array
        longitude of ascending node (degrees)
    omega : (n) float array
        argument of periapsis (degrees)
    M : (n) float array
        Mean anomaly (degrees)

    Returns
    ----------
    rvec, vvec  : tuple of float vectors

    rvec : (n,3) float rray
        Cartesian position vector
    vvec : (n,3) float array
        Cartesian velocity vector
    """
    vecfunc = np.vectorize(el2xv_one, signature='(),(),(),(),(),(),()->(3),(3)')
    return vecfunc(mu, a, ecc, inc, Omega, omega, M)

def xv2el_one(mu,rvec,vvec):
    """
    Converts from cartesian position and velocity values to orbital elements

    Parameters
    ----------
    mu : float
        Central body gravitational constant
    rvec : (3) float array
        Cartesian position vector
    vvec : (3) float array
        Cartesian velocity vector

    Returns
    ----------
    a, ecc, inc, Omega, omega, M, varpi, f, lam : tuple of floats

    a : float
        semimajor axis
    ecc : float
        eccentricity
    inc : float
        inclination (degrees)
    Omega : float
        longitude of ascending node (degrees)
    omega : float
        argument of periapsis (degrees)
    M : float
        mean anomaly (degrees)
    varpi : flaot
        longitude of periapsis (degrees)
    f : float
        true anomaly (degrees)
    lam : float
        mean longitude (degrees)

    """

    rmag = np.linalg.norm(rvec)
    vmag2 = np.vdot(vvec,vvec)
    h = np.cross(rvec,vvec)
    hmag = np.linalg.norm(h)

    rdot = np.sign(np.vdot(rvec,vvec)) * np.sqrt(vmag2 - (hmag / rmag)**2)

    a = 1.0/(2.0 / rmag - vmag2/mu)
    ecc = np.sqrt(1 - hmag**2 / (mu * a))
    inc = np.arccos(h[2]/hmag)

    goodinc = np.abs(inc) > np.finfo(np.float64).tiny
    sO = np.where(goodinc,  np.sign(h[2]) * h[0] / (hmag * np.sin(inc)),0.0)
    cO = np.where(goodinc, -np.sign(h[2]) * h[1] / (hmag * np.sin(inc)),1.0)

    Omega = np.arctan2(sO, cO)

    sof = np.where(goodinc,rvec[2] / (rmag * np.sin(inc)), rvec[1]/rmag)
    cof = np.where(goodinc,(rvec[0] / rmag + np.sin(Omega) * sof * np.cos(inc)) / np.cos(Omega), rvec[0]/rmag)

    of = np.arctan2(sof,cof)

    goodecc = ecc > np.finfo(np.float64).tiny
    sf = np.where(goodecc,a * (1.0 - ecc**2)  * rdot / (hmag * ecc), 0.0)
    cf = np.where(goodecc,(1.0 / ecc) * (a * (1.0 - ecc**2) / rmag - 1.0), 1.0 )

    f = np.arctan2(sf, cf)

    omega = of - f

    varpi = Omega + omega

    # Compute eccentric anomaly & mean anomaly in order to get mean longitude
    E = np.where(ecc > np.finfo(np.float64).tiny, np.arccos(-(rmag - a) / (a * ecc)), 0)
    if np.sign(np.vdot(rvec, vvec)) < 0.0:
        E = 2 * np.pi - E

    M = E - ecc * np.sin(E)
    lam = M + varpi

    return a, ecc, np.rad2deg(inc), np.rad2deg(Omega), np.rad2deg(omega), np.rad2deg(M), np.rad2deg(varpi), np.rad2deg(f), np.rad2deg(lam)

def xv2el_vec(mu, rvec, vvec):
    """
    Vectorized call to xv2el_one.

    Parameters
    ----------
    mu : float
        Central body gravitational constant
    rvec : (n,3) float array
        Cartesian position vector
    vvec : (n,3) float array
        Cartesian velocity vector

    Returns
    ----------
    a, ecc, inc, Omega, omega, M, varpi, f, lam : tuple of float arrays

    a : (n) float array
        semimajor axis
    ecc : (n) float array
        eccentricity
    inc : (n) float array
        inclination (degrees)
    Omega : (n) float array
        longitude of ascending node (degrees)
    omega : (n) float array
        argument of periapsis (degrees)
    M : (n) float array
        mean anomaly (degrees)
    varpi : (n) flaot array
        longitude of periapsis (degrees)
    f : (n) float array
        true anomaly (degrees)
    lam : (n) float array
        mean longitude (degrees)

    """

    vecfunc = np.vectorize(xv2el_one, signature='(),(3),(3)->(),(),(),(),(),(),(),(),()')
    return vecfunc(mu, rvec, vvec)


"""
Copyright 2022 - David Minton, Carlisle Wishard, Jennifer Pouplin, Jake Elliott, & Dana Singh
This file is part of Swiftest.
Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Swiftest. 
If not, see: https://www.gnu.org/licenses. 
"""
from __future__ import annotations

import swiftest
import numpy as np
import numpy.typing as npt
from astroquery.jplhorizons import Horizons
import astropy.units as u
from astropy.coordinates import SkyCoord
import datetime
import xarray as xr
from typing import (
    Literal,
    Dict,
    List,
    Any
)

def horizons_get_physical_properties(altid,**kwargs):
    """
    Parses the raw output from JPL Horizons in order to extract physical properties of a body if they exist
    

    Parameters
    ----------
    altid : list of str
        List of ids to use for Horizons query
    **kwargs: Any
            Additional keyword arguments to pass to the query method (see https://astroquery.readthedocs.io/en/latest/jplhorizons/jplhorizons.html)


    Returns
    -------
    MSun_over_Mpl : float
        The ratio of MSun/M of the body
    radius : float
        The radius of the body in m
    rot: (3) float vector
        The rotation rate vector oriented toward the north pole
    """

    def get_Gmass(raw_response):
        GM = [s for s in raw_response.split('\n') if 'GM' in s 
            and 'Rel' not in s 
            and 'GMT' not in s
            and 'ANGMOM' not in s]
        if len(GM) == 0:
            return None
        GM = GM[0]
        if len(GM) > 1:
            GM = GM.split('GM')[1]
            if len(GM) > 1:
                GM = GM.split('=')
                if len(GM) > 1:
                    GM = GM[1].strip().split(' ')[0].split('+')[0]
        try:
            GM = float(GM)
        except:
            GM = None
        return GM

    def get_radius(raw_response):

        radius = [s for s in raw_response.split('\n') if 'mean radius' in s.lower() or 'RAD' in s or 'Radius (km)' in s]
        if len(radius) == 0:
            return None
        radius = radius[0]
        if "Radius (km)" in radius:
            radius = radius.split("Radius (km)")[1].strip(' =').split('+')[0].split()[0].strip()
        elif "R_eff" in radius: # Some small bodies list an effective radius 
            radius = radius.split('R_eff')[1].strip(' =').split('+')[0].split()[0].strip()
        elif "RAD" in radius: # Some small bodies list the radius like this
            radius = radius.split('RAD')[1].strip(' =').split('+')[0].split()[0].strip()
            if 'x' in radius: # Triaxial ellipsoid bodies like Haumea may have multiple dimensions which need to be averaged
                radius = radius.split('x')
                try:
                    for i,v in enumerate(radius):
                        radius[i] = float(v.split()[0].strip())
                    radius = np.average(radius)
                except:
                    radius = None
        else: # Handles most major bodies
            radius = radius.split('=')[1].strip().split('+')[0].split()[0].strip() 
        try:
            radius = float(radius)
        except:
            radius = None
        return radius

    def get_rotrate(raw_response):
        raw_response=jpl.raw_response
        rotrate = [s for s in raw_response.split('\n') if 'rot. rat' in s.lower()]
        if len(rotrate) == 0:
            rotrate = [s for s in raw_response.split('\n') if 'ROTPER' in s.upper()] # Try the small body version
            if len(rotrate) > 0:
                rotrate = rotrate[0].split('ROTPER=')[1].strip()
                try:
                   rotrate = 2*np.pi / (float(rotrate) * 3600)
                except:
                   rotrate = None
            else:
                if "Synchronous" in raw_response: # Satellites have this:
                    rotrate = [s for s in raw_response.split('\n') if 'Orbital period' in s][0]
                    rotrate = rotrate.split('Orbital period')[1].replace('~',' ').replace('d',' ').replace('=',' ').strip()
                    rotrate = 2*np.pi / (float(rotrate) *  swiftest.JD2S) 
                else:   
                    rotrate = None
        else:
            rotrate = rotrate[0].lower().split('rot. rat')[1].split('=')[1].strip().split('  ')[0].strip()
            try:
                rotrate = float(rotrate)
            except:
                rotrate = None
        return rotrate

    def get_rotpole(jpl):
        RA = jpl.ephemerides()['NPole_RA'][0]
        DEC = jpl.ephemerides()['NPole_DEC'][0]

        if np.ma.is_masked(RA) or np.ma.is_masked(DEC):
            return np.array([0.0,0.0,1.0])

        rotpole = SkyCoord(ra=RA * u.degree, dec=DEC * u.degree).cartesian
        return np.array([rotpole.x.value, rotpole.y.value, rotpole.z.value])
    
    if type(altid) != list:
        altid = [altid]

    for id in altid:
        jpl,idlist,namelist = horizons_query(id=id,ephemerides_start_date='2023-07-26',verbose=False,**kwargs)
        Rpl = get_radius(jpl.raw_response) 
        if Rpl is not None:
            Rpl *= 1e3
            break
        
    Gmass = get_Gmass(jpl.raw_response)
    if Rpl is None or Gmass is None:
        rot = np.full(3,np.nan) 
    else:
        print(f"Physical properties found for {namelist[0]}") 
        Gmass *= 1e9  # JPL returns GM in units of km**3 / s**2, so convert to SI

        rotrate = get_rotrate(jpl.raw_response)
        if rotrate is None:
            rotrate = 0.0
        else:
            rotrate = np.rad2deg(rotrate)

        rotpole = get_rotpole(jpl)
        rot = rotpole*rotrate
            
    return Gmass,Rpl,rot


def horizons_query(id, ephemerides_start_date, exclude_spacecraft=True, verbose=False,**kwargs):
    """
    Queries JPL/Horizons for a body matching the id. If one is found, a HorizonsClass object is returned for the first object that
    matches the passed id string. If more than one match is found, a list of alternate ids is also returned. If no object is found
    then None is returned.

    Parameters
    ----------
    id : string
        A string identifying which body is requested from JPL/Horizons
    ephemerides_start_date : string
        Date to use when obtaining the ephemerides in the format YYYY-MM-DD.
    exclude_spacecraft: bool (optional) - Default True
        Indicate whether spacecraft ids should be exluded from the alternate id list
    verbose: bool (optional) - Default True
        Indicate whether to print messages about the query or not

    Returns
    -------
    jpl: HorizonsClass | None
        An astroquery.jplhorizons HorizonsClass object. Or None if no match was found.
    altid: string list | None
        A list of alternate ids if more than one object matches the list
        
    """
    
    def get_altid(errstr,exclude_spacecraft=True):
        """
        Parses the error message returned from a failed Horizons query. If the query failed because of an ambiguous id, then it will
        return a list of id values that could possibly match the query.
        not found

        Parameters
        ----------
        raw_response : string
            Raw response from the JPL Horizons query

        Returns
        -------
        MSun_over_Mpl : float
        """    
        if "ID" in errstr:
            altid = errstr.split('ID')[1]
            altid = altid.split('\n')[2:-1]
            altname = []
            for n,v in enumerate(altid):
                altid[n] = v.strip().split(' ')[0]
                altname.append(' '.join(v.strip().split(' ')[1:]).strip().split('  ')[0])
            if exclude_spacecraft:
                altname = [altname[i] for i,n in enumerate(altid) if int(n) > 0] 
                altid = [n for n in altid if int(n) > 0]
                
            return altid,altname 
        else:
            return None,None
        
        
    # Horizons date time internal variables
    tstart = datetime.date.fromisoformat(ephemerides_start_date)
    tstep = datetime.timedelta(days=1)
    tend = tstart + tstep
    ephemerides_end_date = tend.isoformat()
    ephemerides_step = '1d'
    
    try:
        jpl = Horizons(id=id, location='@sun',
                            epochs={'start': ephemerides_start_date, 'stop': ephemerides_end_date,
                                    'step': ephemerides_step},**kwargs)
        eph=jpl.ephemerides()
        altid = [id]
        altname =[jpl.table['targetname'][0]]
    except Exception as e:
        altid,altname = get_altid(str(e))
        if altid is not None and len(altid) >0: # Return the first matching id
            id = altid[0]
            jpl = Horizons(id=id, location='@sun',
                        epochs={'start': ephemerides_start_date, 'stop': ephemerides_end_date,
                                'step': ephemerides_step})
            eph=jpl.ephemerides()
        else:
            print(f"Could not find {id} in the JPL/Horizons system")
            return None,None,None
    if verbose:
        print(f"Found matching body: {altname[0]} ({altid[0]})") 
        if len(altid) > 1:
            print("    Alternate matching bodies:")
            for i,v in enumerate(altid):
                if i > 0:
                    print(f"    {altname[i]} ({v})")
        
    return jpl,altid,altname
    
    
def solar_system_horizons(name: str,
                          param: Dict,
                          ephemerides_start_date: str,
                          ephemeris_id: str | None = None,
                          **kwargs: Any):
    """
    Initializes a Swiftest dataset containing the major planets of the Solar System at a particular data from JPL/Horizons

    Parameters
    ----------
    name  : string
        Name of body to add to Dataset. If `id` is not supplied, this is also what will be searche for in the JPL/Horizon's database.
        The first matching body is found (for major planets, this is the barycenter of a planet-satellite system)
    param : dict
        Swiftest paramuration parameters. This method uses the unit conversion factors to convert from JPL's AU-day system into the system specified in the param file
    ephemerides_start_date : string
        Date to use when obtaining the ephemerides in the format YYYY-MM-DD.
    ephemeris_id : string (optional)
        If passed, this is passed to Horizons instead of `name`. This can be used to find a more precise body than given by `name`. 
    **kwargs: Any
            Additional keyword arguments to pass to the query method (see https://astroquery.readthedocs.io/en/latest/jplhorizons/jplhorizons.html)

    Returns
    -------
    ds : xarray dataset
        Initial conditions of body formatted for Swiftest
    
    Notes
    --------
    When passing `name` == "Earth" or `name` == "Pluto", it a body is generated that has initial conditions matching the system
    barycenter and mass equal to the sum of Earth+Moon or Pluto+Charon. To obtain initial conditions for either Earth or Pluto alone,
    pass `ephemeris_id` == "399" for Earth or `id` == "999" for Pluto. 
    """
    
    planetIpz = { # Only the polar moments of inertia are used currently. Where the quantity is unkown, we just use the value of a sphere = 0.4
        'Sun' : np.longdouble(0.070),
        'Mercury' : np.longdouble(0.346),
        'Venus': np.longdouble(0.4),
        'Earth': np.longdouble(0.3307),
        'Mars': np.longdouble(0.3644),
        'Jupiter': np.longdouble(0.2756),
        'Saturn': np.longdouble(0.22),
        'Uranus': np.longdouble(0.23),
        'Neptune': np.longdouble(0.23),
        'Pluto': np.longdouble(0.4)
        }

    # Unit conversion factors
    DCONV = swiftest.AU2M / param['DU2M']
    VCONV = (swiftest.AU2M / swiftest.JD2S) / (param['DU2M'] / param['TU2S'])
    THIRDLONG = np.longdouble(1.0) / np.longdouble(3.0)
    
    # Central body value vectors
    GMcb = swiftest.GMSun * param['TU2S'] ** 2 / param['DU2M'] ** 3
    Rcb = swiftest.RSun / param['DU2M']
    J2RP2 = swiftest.J2Sun * (swiftest.RSun / param['DU2M']) ** 2
    J4RP4 = swiftest.J4Sun * (swiftest.RSun / param['DU2M']) ** 4
    
    rotcb = swiftest.rotSun * param['TU2S'] 
    rotcb = np.array([rotcb.x.value, rotcb.y.value, rotcb.z.value])
    Ipsun = np.array([0.0, 0.0, planetIpz['Sun']])

    param_tmp = param
    param_tmp['OUT_FORM'] = 'XVEL'

    rh = np.full(3,np.nan)
    vh = np.full(3,np.nan)
    a = None
    e = None
    inc = None
    capom = None
    omega = None
    capm = None
    Ip = np.full(3,np.nan)
    rot = np.full(3,np.nan)
    rhill = None
    Gmass = None
    Rpl = None
    J2 = None
    J4 = None

    if name == "Sun" or ephemeris_id == "0": # Create central body
        print("Creating the Sun as a central body")
        Gmass = GMcb
        Rpl = Rcb
        J2 = J2RP2
        J4 = J4RP4
        if param['ROTATION']:
            Ip = Ipsun
            rot = rotcb
    else: # Fetch solar system ephemerides from Horizons
        if ephemeris_id is None:
            ephemeris_id = name
            
        print(f"Fetching ephemerides data for {ephemeris_id} from JPL/Horizons")
        
        jpl,altid,altname = horizons_query(ephemeris_id,ephemerides_start_date,**kwargs)
        if jpl is not None:
            print(f"Found ephemerides data for {altname[0]} ({altid[0]}) from JPL/Horizons")
            if name == None:
                name = altname[0]
        else:
            return None
        
        if param['IN_FORM'] == 'XV':
            rx = jpl.vectors()['x'][0] * DCONV
            ry = jpl.vectors()['y'][0] * DCONV
            rz = jpl.vectors()['z'][0] * DCONV
            vx = jpl.vectors()['vx'][0] * VCONV
            vy = jpl.vectors()['vy'][0] * VCONV
            vz = jpl.vectors()['vz'][0] * VCONV

            rh = np.array([rx,ry,rz])
            vh = np.array([vx,vy,vz])
        elif param['IN_FORM'] == 'EL':
            a = jpl.elements()['a'][0] * DCONV
            e = jpl.elements()['e'][0]
            inc = jpl.elements()['incl'][0]
            capom = jpl.elements()['Omega'][0]
            omega = jpl.elements()['w'][0]
            capm = jpl.elements()['M'][0]

        Gmass,Rpl,rot = horizons_get_physical_properties(altid,**kwargs)
        # If the user inputs "Earth" or Pluto, then the Earth-Moon or Pluto-Charon barycenter and combined mass is used. 
        # To use the Earth or Pluto alone, simply pass "399" or "999", respectively to name
        if ephemeris_id == "Earth":
            print("Combining mass of Earth and the Moon")
            Gmass_moon,tmp,tmp = horizons_get_physical_properties(["301"],**kwargs)
            Gmass += Gmass_moon
        elif ephemeris_id == "Pluto":
            print("Combining mass of Pluto and Charon")
            Gmass_charon,tmp,tmp = horizons_get_physical_properties(["901"],**kwargs)
            Gmass += Gmass_charon 
        
        if Gmass is not None:
            # Convert from SI to system units
            Gmass *= param['TU2S'] ** 2 / param['DU2M'] ** 3
            
            if param['CHK_CLOSE']:
                Rpl /= param['DU2M']

            # Generate planet value vectors
            if (param['RHILL_PRESENT']):
                rhill = jpl.elements()['a'][0] * DCONV * (3 * Gmass / GMcb) ** (-THIRDLONG)

            if (param['ROTATION']):
                rot *= param['TU2S']
                if name in planetIpz:
                    Ip = np.array([0.0, 0.0, planetIpz[name]])
                else:
                    Ip = np.array([0.4, 0.4, 0.4])
        else:
            Gmass = None

    # Only the Sun gets assigned its own special id for now. All other ids will be sorted later
    if name == "Sun":
        id = 0
    else:
        id = -1

    return id,name,a,e,inc,capom,omega,capm,rh,vh,Gmass,Rpl,rhill,Ip,rot,J2,J4

def vec2xr(param: Dict, **kwargs: Any):
    """
    Converts and stores the variables of all bodies in an xarray dataset.

    Parameters
    ----------
    param : dict
        Swiftest simulation parameters.
    name : str or array-like of str, optional
        Name or names of Bodies. If none passed, name will be "Body<id>"
    id : int or array-like of int, optional
        Unique id values. If not passed, an id will be assigned in ascending order starting from the pre-existing
        Dataset ids.
    a : float or array-like of float, optional
        semimajor axis for param['IN_FORM'] == "EL"
    e : float or array-like of float, optional
        eccentricity  for param['IN_FORM'] == "EL"
    inc : float or array-like of float, optional
        inclination for param['IN_FORM'] == "EL"
    capom : float or array-like of float, optional
        longitude of periapsis for param['IN_FORM'] == "EL"
    omega : float or array-like of float, optional
        argument of periapsis for param['IN_FORM'] == "EL"
    capm : float or array-like of float, optional
        mean anomaly for param['IN_FORM'] == "EL"
    rh : (n,3) array-like of float, optional
        Position vector array.
    vh : (n,3) array-like of float, optional
        Velocity vector array. 
    Gmass : float or array-like of float, optional
        G*mass values if these are massive bodies (only one of mass or Gmass can be passed)
    radius : float or array-like of float, optional
        Radius values if these are massive bodies
    rhill : float or array-like of float, optional
        Hill's radius values if these are massive bodies
    rot:  (n,3) array-like of float, optional
        Rotation rate vectors if these are massive bodies with rotation enabled in deg/TU
    Ip: (n,3) array-like of flaot, optional
        Principal axes moments of inertia vectors if these are massive bodies with rotation enabled. This can be used
        instead of passing Ip1, Ip2, and Ip3 separately
    time : array of floats
        Time at start of simulation
    Returns
    -------
    ds : xarray dataset
    """
    scalar_dims = ['id']
    vector_dims = ['id','space']
    space_coords = np.array(["x","y","z"])

    vector_vars = ["rh","vh","Ip","rot"]
    scalar_vars = ["name","a","e","inc","capom","omega","capm","Gmass","radius","rhill","j2rp2","j4rp4"]
    time_vars =  ["rh","vh","Ip","rot","a","e","inc","capom","omega","capm","Gmass","radius","rhill","j2rp2","j4rp4"]

    # Check for valid keyword arguments
    kwargs = {k:kwargs[k] for k,v in kwargs.items() if v is not None}
  
    if "ROTATION" in param and param['ROTATION'] == True: 
        if "rot" not in kwargs and "Gmass" in kwargs:
            kwargs['rot'] = np.zeros((len(kwargs['Gmass']),3))
        if "Ip" not in kwargs and "Gmass" in kwargs:
            kwargs['Ip'] = np.full((len(kwargs['Gmass']),3), 0.4)

    if "time" not in kwargs:
        kwargs["time"] = np.array([0.0])

    valid_arguments = vector_vars + scalar_vars + ['time','id']

    kwargs = {k:v for k,v in kwargs.items() if k in valid_arguments}

    data_vars = {k:(scalar_dims,v) for k,v in kwargs.items() if k in scalar_vars}
    data_vars.update({k:(vector_dims,v) for k,v in kwargs.items() if k in vector_vars})
    ds = xr.Dataset(data_vars=data_vars,
                    coords={
                        "id":(["id"],kwargs['id']),
                        "space":(["space"],space_coords),
                    }
                    )
    time_vars = [v for v in time_vars if v in ds]
    for v in time_vars:
        ds[v] = ds[v].expand_dims({"time":1}).assign_coords({"time": kwargs['time']})

    return ds


#!/usr/bin/env python3

"""
 Copyright 2023 - David Minton, Carlisle Wishard, Jennifer Pouplin, Jake Elliott, & Dana Singh
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""

"""
Generates and runs two sets of Swiftest input files from initial conditions with the helio integrator. All simulation 
outputs for the general relativity run are stored in the /gr subdirectory while all simulation outputs for the run 
without general reelativity are stored in the /nogr subdirectory.

Input
------
None.

Output
------
helio_gr_mercury_precession.png : Portable Network Graphic file depicting the precession of Mercury's perihelion over time
                                  with data sourced from the JPL Horizons database, Swiftest run with general relativity, 
                                  and Swiftest run without general relativity.
Two subdirectories:
gr/
nogr/

Each subdirecotry contains:
collisions.log               : A NetCDF file containing the collision output.
data.nc                      : A NetCDF file containing the simulation output.
init_cond.nc                 : A NetCDF file containing the initial conditions for the simulation.
param.00...0.in              : A series of parameter input files containing the parameters for the simulation at every output stage.
param.in                     : An ASCII file containing the parameters for the simulation.
param.restart.in             : An ASCII file containing the parameters for the simulation at the last output. 
swiftest.log                 : An ASCII file containing the information on the status of the simulation as it runs.
"""

import swiftest
from astroquery.jplhorizons import Horizons
import datetime
import numpy as np
import matplotlib.pyplot as plt

# Initialize the simulation object as a variable. Define the directory in which the output will be placed.
sim_gr = swiftest.Simulation(simdir="gr")
# Add the modern planets and the Sun using the JPL Horizons Database.
sim_gr.add_solar_system_body(["Sun","Mercury","Venus","Earth","Mars","Jupiter","Saturn","Uranus","Neptune"])

# Initialize the simulation object as a variable. Define the directory in which the output will be placed.
sim_nogr = swiftest.Simulation(simdir="nogr")
# Add the modern planets and the Sun using the JPL Horizons Database.
sim_nogr.add_solar_system_body(["Sun","Mercury","Venus","Earth","Mars","Jupiter","Saturn","Uranus","Neptune"])

# Define a set of arguments that apply to both runs. For a list of possible arguments, see the User Manual.
run_args = {"tstop":1000.0, "dt":0.005, "tstep_out":10.0, "dump_cadence": 0,"integrator":"helio"}

# Run both simulations.
sim_gr.run(**run_args,general_relativity=True)
sim_nogr.run(**run_args,general_relativity=False)

# Get the start and end date of the simulation so we can compare with the real solar system.
start_date = sim_gr.ephemeris_date
tstop_d = sim_gr.param['TSTOP'] * sim_gr.param['TU2S'] / swiftest.JD2S

stop_date = (datetime.datetime.fromisoformat(start_date) + datetime.timedelta(days=tstop_d)).isoformat()

#Get the ephemerides of Mercury for the same timeframe as the simulation.
obj = Horizons(id='1', location='@sun',
               epochs={'start':start_date, 'stop':stop_date,
                       'step':'10y'})
el = obj.elements()
t = (el['datetime_jd']-el['datetime_jd'][0]) / 365.25
varpi_obs = el['w'] + el['Omega']

varpisim_gr= sim_gr.data['varpi'].sel(name="Mercury")
varpisim_nogr= sim_nogr.data['varpi'].sel(name="Mercury")
tsim = sim_gr.data['time']

dvarpi_gr = np.diff(varpisim_gr)  * 3600 * 100 / run_args['tstep_out']
dvarpi_nogr = np.diff(varpisim_nogr)  * 3600 * 100 / run_args['tstep_out']
dvarpi_obs = np.diff(varpi_obs) / np.diff(t) * 3600 * 100

# Plot of the data and save the output plot.
fig, ax = plt.subplots()

ax.plot(t, varpi_obs, label="JPL Horizons",linewidth=2.5)
ax.plot(tsim, varpisim_gr, label="Swiftest Helio GR",linewidth=1.5)
ax.plot(tsim, varpisim_nogr, label="Swiftest Helio No GR",linewidth=1.5)
ax.set_xlabel('Time (y)')
ax.set_ylabel('Mercury $\\varpi$ (deg)')
ax.legend()
plt.savefig("helio_gr_mercury_precession.png",dpi=300)

# Print the data to the terminal.
print('Mean precession rate for Mercury long. peri. (arcsec/100 y)')
print(f'JPL Horizons         : {np.mean(dvarpi_obs)}')
print(f'Swiftest No GR       : {np.mean(dvarpi_nogr)}')
print(f'Swiftest GR          : {np.mean(dvarpi_gr)}')
print(f'Obs - Swiftest GR    : {np.mean(dvarpi_obs - dvarpi_gr)}')
print(f'Obs - Swiftest No GR : {np.mean(dvarpi_obs - dvarpi_nogr)}')



#!/usr/bin/env python3

"""
 Copyright 2023 - David Minton, Carlisle Wishard, Jennifer Pouplin, Jake Elliott, & Dana Singh
 This file is part of Swiftest.
 Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
 of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 You should have received a copy of the GNU General Public License along with Swiftest. 
 If not, see: https://www.gnu.org/licenses. 
"""

"""
Generates and runs a set of Swiftest input files from initial conditions with the SyMBA integrator. All simulation 
outputs are stored in the /simdata subdirectory.

Input
------
None.

Output
------
collisions.log   : An ASCII file containing the information of any collisional events that occured.
collisions.nc    : A NetCDF file containing the collision output.
data.nc          : A NetCDF file containing the simulation output.
init_cond.nc     : A NetCDF file containing the initial conditions for the simulation.
param.00...0.in  : A series of parameter input files containing the parameters for the simulation at every output stage.
param.in         : An ASCII file containing the inital parameters for the simulation.
param.restart.in : An ASCII file containing the parameters for the simulation at the last output. 
swiftest.log     : An ASCII file containing the information on the status of the simulation as it runs.
"""

import swiftest
import numpy as np
from numpy.random import default_rng

# Initialize the simulation object as a variable. Arguments may be defined here or through the sim.run() method.
#sim = swiftest.Simulation(fragmentation=True, minimum_fragment_mass = 2.5e-11, mtiny=2.5e-8)
sim = swiftest.Simulation()
sim.clean()
rng = default_rng(seed=123)

# Add the modern planets and the Sun using the JPL Horizons Database.
sim.add_solar_system_body(["Sun","Mercury","Venus","Earth","Mars","Jupiter","Saturn","Uranus","Neptune"])

# Add in some main belt asteroids
sim.add_solar_system_body(name=["Ceres","Vesta","Pallas","Hygiea"],id_type="smallbody")

# Add in some big KBOs
sim.add_solar_system_body(name=["Pluto","Eris","Sedna","Haumea","Makemake","Quaoar","Orcus","Gonggong","Salacia"])

# Add 5 user-defined massive bodies.
npl         = 5
density_pl  = 3000.0 / (sim.param['MU2KG'] / sim.param['DU2M'] ** 3)

name_pl     = ["SemiBody_01", "SemiBody_02", "SemiBody_03", "SemiBody_04", "SemiBody_05"]
a_pl        = rng.uniform(0.3, 1.5, npl)
e_pl        = rng.uniform(0.0, 0.2, npl)
inc_pl      = rng.uniform(0.0, 10, npl)
capom_pl    = rng.uniform(0.0, 360.0, npl)
omega_pl    = rng.uniform(0.0, 360.0, npl)
capm_pl     = rng.uniform(0.0, 360.0, npl)
M_pl        = np.array([6e20, 8e20, 1e21, 3e21, 5e21]) * sim.KG2MU
R_pl        = np.full(npl, (3 * M_pl/ (4 * np.pi * density_pl)) ** (1.0 / 3.0))
Ip_pl       = np.full((npl,3),0.4,)
rot_pl      = np.zeros((npl,3))
mtiny       = 1.1 * np.max(M_pl)

sim.add_body(name=name_pl, a=a_pl, e=e_pl, inc=inc_pl, capom=capom_pl, omega=omega_pl, capm=capm_pl, mass=M_pl, radius=R_pl,  Ip=Ip_pl, rot=rot_pl)

# Add 10 user-defined test particles.
ntp = 10

name_tp     = ["TestParticle_01", "TestParticle_02", "TestParticle_03", "TestParticle_04", "TestParticle_05", "TestParticle_06", "TestParticle_07", "TestParticle_08", "TestParticle_09", "TestParticle_10"]
a_tp        = rng.uniform(0.3, 1.5, ntp)
e_tp        = rng.uniform(0.0, 0.2, ntp)
inc_tp      = rng.uniform(0.0, 10, ntp)
capom_tp    = rng.uniform(0.0, 360.0, ntp)
omega_tp    = rng.uniform(0.0, 360.0, ntp)
capm_tp     = rng.uniform(0.0, 360.0, ntp)

sim.add_body(name=name_tp, a=a_tp, e=e_tp, inc=inc_tp, capom=capom_tp, omega=omega_tp, capm=capm_tp)
sim.set_parameter(tstart=0.0, tstop=1.0e3, dt=0.01, istep_out=100, dump_cadence=0, compute_conservation_values=True, mtiny=mtiny)
# Display the run configuration parameters.
sim.get_parameter()
sim.save()

# Run the simulation. Arguments may be defined here or thorugh the swiftest.Simulation() method.
sim.run()



# cython: language_level=3, c_string_type=unicode, c_string_encoding=ascii

cdef extern from "_bindings.h":
    void bindings_c_driver(char* integrator, char* param_file_name, char* display_style) noexcept nogil

def driver(integrator, param_file_name, display_style):
    b_integrator = bytes(integrator,'ascii') + b'\x00'
    b_param_file_name = bytes(param_file_name,'ascii') + b'\x00'
    b_display_style = bytes(display_style,'ascii') + b'\x00'

    cdef:
        char* c_integrator = b_integrator 
        char* c_param_file_name = b_param_file_name 
        char* c_display_style = b_display_style 

    try:
        with nogil:
            bindings_c_driver(c_integrator, c_param_file_name, c_display_style)
    except:
        raise Warning("The Swiftest driver did not terminate normally")

    return


#!/usr/bin/env python3
import numpy as np
import swiftest

tstart = 0.0
dt = 1.0
tstop = 365.25e2
tstep_out = 100*dt

sim = swiftest.Simulation(simdir="swiftest_sim",init_cond_format="XV",output_format="XV",general_relativity=False, integrator="RMVS", rhill_present=True, MU="Msun", DU="AU", TU="d")
sim.clean()
sim.add_solar_system_body(["Sun","Mercury","Venus","Earth","Mars","Jupiter","Saturn","Uranus","Neptune"])

plname = sim.init_cond['name'].where(sim.init_cond['name'] != "Sun", drop=True)
pl = sim.init_cond.sel(name=plname)

for i,n in enumerate(pl.name):
    pli = pl.sel(name=n)
    rstart = 2 * pli['radius'].data[0]  # Start the test particles at a multiple of the planet radius away
    vstart = 1.5 * np.sqrt(2 * pli['Gmass'].data[0]  / rstart)  # Start the test particle velocities at a multiple of the escape speed
    rstart_vec = np.array([rstart / np.sqrt(2.0), rstart / np.sqrt(2.0), 0.0])
    vstart_vec = np.array([vstart, 0.0, 0.0])
    rp = pli['rh'].data[0]
    vp = pli['vh'].data[0]
    sim.add_body(name=[f"TestParticle{100+i}",f"TestParticle{200+i}"],rh=[rp+rstart_vec, rp-rstart_vec],vh=[vp+vstart_vec, vp-vstart_vec])


sim.set_parameter(tstart=tstart, tstop=tstop, dt=dt, tstep_out=tstep_out, dump_cadence=0)
sim.save()

sim.set_parameter(simdir="swifter_sim",codename="Swifter",init_cond_file_type="ASCII",output_file_type="REAL8")
sim.save()



"""
Copyright 2022 - David Minton, Carlisle Wishard, Jennifer Pouplin, Jake Elliott, & Dana Singh
This file is part of Swiftest.
Swiftest is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License 
as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Swiftest is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with Swiftest. 
If not, see: https://www.gnu.org/licenses. 
"""

from .constants import *
from .simulation_class import Simulation